@@ -38,12 +38,15 @@ package Mail::SpamAssassin::DnsResolver;
 use strict;
 use warnings;
 use bytes;
+use re 'taint';
 
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Logger;
 
+use Socket;
 use IO::Socket::INET;
 use Errno qw(EADDRINUSE EACCES);
+use Time::HiRes qw(time);
 
 use constant HAS_SOCKET_INET6 => eval { require IO::Socket::INET6; };
 
@@ -90,7 +93,7 @@ sub load_resolver {
                                          Proto     => 'udp',
                                          );
       if ($sock6) {
-        $sock6->close();
+        $sock6->close()  or die "error closing inet6 socket: $!";
         1;
       }
     } ||
@@ -102,7 +105,7 @@ sub load_resolver {
                                          Proto     => 'udp',
                                          );
       if ($sock6) {
-        $sock6->close();
+        $sock6->close()  or die "error closing inet4 socket: $!";
         1;
       }
     };
@@ -127,13 +130,16 @@ sub load_resolver {
       $self->{res}->persistent_udp(0);  # bug 3997
     }
     1;
-  };   #  or warn "dns: eval failed: $@ $!\n";
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+    dbg("dns: eval failed: $eval_stat");
+  };
 
   dbg("dns: no ipv6") if $force_ipv4;
-  dbg("dns: is Net::DNS::Resolver available? " .
-       ($self->{no_resolver} ? "no" : "yes"));
+  dbg("dns: is Net::DNS::Resolver available? %s",
+      $self->{no_resolver} ? "no" : "yes" );
   if (!$self->{no_resolver} && defined $Net::DNS::VERSION) {
-    dbg("dns: Net::DNS version: ".$Net::DNS::VERSION);
+    dbg("dns: Net::DNS version: %s", $Net::DNS::VERSION);
   }
 
   return (!$self->{no_resolver});
@@ -175,7 +181,9 @@ sub connect_sock {
 
   return if $self->{no_resolver};
 
-  $self->{sock}->close() if $self->{sock};
+  if ($self->{sock}) {
+    $self->{sock}->close()  or die "error closing socket: $!";
+  }
   my $sock;
   my $errno;
 
@@ -196,7 +204,7 @@ sub connect_sock {
     $srcaddr = "0.0.0.0";
   }
 
-  dbg("dns: name server: $ns, LocalAddr: $srcaddr");
+  dbg("dns: name server: %s, LocalAddr: %s", $ns,$srcaddr);
 
   # find next available unprivileged port (1024 - 65535)
   # starting at a random value to spread out use of ports
@@ -222,17 +230,37 @@ sub connect_sock {
     if (defined $sock) {  # ok, got it
       last;
     } elsif ($! == EADDRINUSE || $! == EACCES) {  # in use, let's try another source port
-      dbg("dns: UDP port $lport already in use, trying another port");
+      dbg("dns: UDP port %s already in use, trying another port", $lport);
     } else {
-      warn "Error creating a DNS resolver socket: $errno";
+      warn "error creating a DNS resolver socket: $errno";
       goto no_sock;
     }
   }
   if (!defined $sock) {
-    warn "Can't create a DNS resolver socket: $errno";
+    warn "cannot create a DNS resolver socket: $errno";
     goto no_sock;
   }
 
+  eval {
+    my($bufsiz,$newbufsiz);
+    $bufsiz = $sock->sockopt(Socket::SO_RCVBUF)
+      or die "cannot get a resolver socket rx buffer size: $!";
+    if ($bufsiz >= 32*1024) {
+      dbg("dns: resolver socket rx buffer size is %d bytes", $bufsiz);
+    } else {
+      $sock->sockopt(Socket::SO_RCVBUF, 32*1024)
+        or die "cannot set a resolver socket rx buffer size: $!";
+      $newbufsiz = $sock->sockopt(Socket::SO_RCVBUF)
+        or die "cannot get a resolver socket rx buffer size: $!";
+      dbg("dns: resolver socket rx buffer size changed from %d to %d bytes",
+          $bufsiz, $newbufsiz);
+    }
+    1;
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+    info("dns: socket buffer size error: $eval_stat");
+  };
+
   $self->{sock} = $sock;
   $self->{sock_as_vec} = $self->fhs_to_vec($self->{sock});
   return;
@@ -292,22 +320,24 @@ sub new_dns_packet {
     $packet = Net::DNS::Packet->new($host, $type, $class);
 
     # a bit noisy, so commented by default...
-    #dbg("dns: new DNS packet time=".time()." host=$host type=$type id=".$packet->id);
-  };
-
-  if ($@) {
+    #dbg("dns: new DNS packet time=%s host=%s type=%s id=%s",
+    #    time, $host, $type, $packet->id);
+    1;
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
     # this can happen if Net::DNS isn't available -- but in this
     # case this function should never be called!
-    warn "dns: cannot create Net::DNS::Packet, but new_dns_packet() was called: $@ $!";
-  }
+    warn "dns: cannot create Net::DNS::Packet, but new_dns_packet() was called: $eval_stat";
+  };
+
   return $packet;
 }
 
 # Internal function used only in this file
 ## compute an unique ID for a packet to match the query to the reply
 ## It must use only data that is returned unchanged by the nameserver.
-## Argument is a Net::DNS::Packet that has a non-empty question section
-## return is an object that can be used as a hash key
+## Argument is a Net::DNS::Packet that has a non-empty question section,
+## return is an (opaque) string that can be used as a hash key
 sub _packet_id {
   my ($self, $packet) = @_;
   my $header = $packet->header;
@@ -316,13 +346,27 @@ sub _packet_id {
   my $ques = $questions[0];
 
   if (defined $ques) {
-    return join '/', $id, $ques->qname, $ques->qtype, $ques->qclass;
+    # Bug 6232: Net::DNS::Packet::new is not consistent in keeping data in
+    # sections of a packet either as original bytes or presentation-encoded:
+    # creating a query packet as above in new_dns_packet() keeps label in
+    # non-encoded form, yet on parsing an answer packet, its query section
+    # is converted to presentation form by Net::DNS::Question::parse calling
+    # Net::DNS::Packet::dn_expand and Net::DNS::wire2presentation in turn.
+    # Let's undo the effect of the wire2presentation routine here to make
+    # sure the query section of an answer packet matches the query section
+    # in our packet formed by new_dns_packet():
+    #
+    my $qname = $ques->qname;
+    $qname =~ s/\\([0-9]{3}|.)/length($1)==1 ? $1 : chr($1)/gse;
+    return join '/', $id, $qname, $ques->qtype, $ques->qclass;
+
   } else {
     # odd.  this should not happen, but clearly some DNS servers
     # can return something that Net::DNS interprets as having no
     # question section.  Better support it; just return the
     # (safe) ID part, along with a text token indicating that
     # the packet had no question part.
+    #
     return $id . "NO_QUESTION_IN_PACKET";
   }
 }
@@ -338,18 +382,19 @@ sub reference C<$cb> will be called.
 Note that C<$type> and C<$class> may be C<undef>, in which case they
 will default to C<A> and C<IN>, respectively.
 
-The callback sub will be called with two arguments -- the packet that was
-delivered and an id string that fingerprints the query packet and the expected reply.
-It is expected that a closure callback be used, like so:
+The callback sub will be called with three arguments -- the packet that was
+delivered, and an id string that fingerprints the query packet and the expected
+reply. The third argument is a timestamp (Unix time, floating point), captured
+at the time the packet was collected. It is expected that a closure callback
+be used, like so:
 
   my $id = $self->{resolver}->bgsend($host, $type, undef, sub {
-        my $reply = shift;
-        my $reply_id = shift;
+        my ($reply, $reply_id, $timestamp) = @_;
         $self->got_a_reply ($reply, $reply_id);
       });
 
-The callback can ignore the reply as an invalid packet sent to the listening port
-if the reply id does not match the return value from bgsend.
+The callback can ignore the reply as an invalid packet sent to the listening
+port if the reply id does not match the return value from bgsend.
 
 =cut
 
@@ -367,6 +412,7 @@ sub bgsend {
     return;
   }
   my $id = $self->_packet_id($pkt);
+  dbg("dns: providing a callback for id: $id");
   $self->{id_to_callback}->{$id} = $cb;
   return $id;
 }
@@ -384,46 +430,53 @@ sub poll_responses {
   my ($self, $timeout) = @_;
   return if $self->{no_resolver};
   return if !$self->{sock};
+  my $cnt = 0;
 
   my $rin = $self->{sock_as_vec};
   my $rout;
-  my ($nfound, $timeleft) = select($rout=$rin, undef, undef, $timeout);
 
-  if (!defined $nfound || $nfound < 0) {
-    warn "dns: select failed: $!";
-    return;
-  }
-
-  if ($nfound == 0) {
-    return 0;           # nothing's ready yet
-  }
-
-  my $packet = $self->{res}->bgread($self->{sock});
-  my $err = $self->{res}->errorstring;
-
-  if (defined $packet &&
-      defined $packet->header &&
-      defined $packet->question &&
-      defined $packet->answer)
-  {
-    my $id = $self->_packet_id($packet);
-
-    my $cb = delete $self->{id_to_callback}->{$id};
-    if (!$cb) {
-      dbg("dns: no callback for id: $id, ignored; packet: ".
-                    ($packet ? $packet->string : "undef"));
-      return 0;
+  for (;;) {
+    my ($nfound, $timeleft);
+    { my $timer;  # collects timestamp when variable goes out of scope
+      if (!defined($timeout) || $timeout > 0)
+        { $timer = $self->{main}->time_method("poll_dns_idle") }
+      ($nfound, $timeleft) = select($rout=$rin, undef, undef, $timeout);
+    }
+    if (!defined $nfound || $nfound < 0) {
+      warn "dns: select failed: $!";
+      return;
     }
 
-    $cb->($packet, $id);
-    return 1;
-  }
-  else {
-    dbg("dns: no packet! err=$err packet=".
-                    ($packet ? $packet->string : "undef"));
+    my $now = time;
+    $timeout = 0;  # next time around collect whatever is available, then exit
+    last  if $nfound == 0;
+
+    my $packet = $self->{res}->bgread($self->{sock});
+    my $err = $self->{res}->errorstring;
+
+    if (defined $packet &&
+        defined $packet->header &&
+        defined $packet->question &&
+        defined $packet->answer)
+    {
+      my $id = $self->_packet_id($packet);
+
+      my $cb = delete $self->{id_to_callback}->{$id};
+      if (!$cb) {
+        dbg("dns: no callback for id: %s, ignored; packet: %s",
+            $id,  $packet ? $packet->string : "undef" );
+      } else {
+        $cb->($packet, $id, $now);
+        $cnt++;
+      }
+    }
+    else {
+      dbg("dns: no packet! err=%s packet=%s",
+          $err,  $packet ? $packet->string : "undef" );
+    }
   }
 
-  return 0;
+  return $cnt;
 }
 
 ###########################################################################
@@ -457,6 +510,7 @@ sub send {
   my $retries = $self->{retry};
   my $timeout = $retrans;
   my $answerpkt;
+  my $answerpkt_avail = 0;
   for (my $i = 0;
        (($i < $retries) && !defined($answerpkt));
        ++$i, $retrans *= 2, $timeout = $retrans) {
@@ -464,18 +518,18 @@ sub send {
     $timeout = 1 if ($timeout < 1);
     # note nifty use of a closure here.  I love closures ;)
     $self->bgsend($name, $type, $class, sub {
-      $answerpkt = shift;
+      my ($reply, $reply_id, $timestamp) = @_;
+      $answerpkt = $reply; $answerpkt_avail = 1;
     });
 
     my $now = time;
     my $deadline = $now + $timeout;
 
-    while (($now < $deadline) && (!defined($answerpkt))) {
+    while (!$answerpkt_avail) {
+      if ($now >= $deadline) { $self->{send_timed_out} = 1; last }
       $self->poll_responses(1);
-      last if defined $answerpkt;
       $now = time;
     }
-    $self->{send_timed_out} = 1 unless ($now < $deadline);
   }
   return $answerpkt;
 }
@@ -511,7 +565,7 @@ Reset socket when done with it.
 sub finish_socket {
   my ($self) = @_;
   if ($self->{sock}) {
-    $self->{sock}->close();
+    $self->{sock}->close()  or die "error closing socket: $!";
     delete $self->{sock};
   }
 }
@@ -539,8 +593,11 @@ sub fhs_to_vec {
   my $rin = '';
   foreach my $sock (@fhlist) {
     my $fno = fileno($sock);
-    warn "dns: oops! fileno now undef for $sock" unless defined($fno);
-    vec ($rin, $fno, 1) = 1;
+    if (!defined $fno) {
+      warn "dns: oops! fileno now undef for $sock";
+    } else {
+      vec ($rin, $fno, 1) = 1;
+    }
   }
   return $rin;
 }
